// This program is an example of an Open Data Services application. It accepts
// requests from clients and then passes those requests on to an SQL server.
// The results from the SQL server are then sent back to the client.


// data structure to save per user connection
typedef struct remote_dbms {
    LOGINREC *login;    // The SQL Server login structure
    DBPROCESS *dbproc;  // The connection to the remote SQL Server
    short retparams[MAXPARAMS]; // tag return parameters
} REMOTE_DBMS;


// create global variable and mutex handle for serializing connect handler
SRV_PROC *Newsrvproc;       
HANDLE init_remote_mutex = 0;

// function prototypes for ODS event handlers
//
int init_server(SRV_SERVER *server);
int init_remote(SRV_PROC *srvproc);
int exit_remote(SRV_PROC *srvproc);
int attn_handler(SRV_PROC *srvproc);
int rpc_execute(SRV_PROC *srvproc);
int lang_execute(SRV_PROC *srvproc);


// utility function prototypes

int handle_results(DBPROCESS *rmtproc, SRV_PROC *srvproc);
int remotemsgs(DBPROCESS *dbproc, DBINT msgno, DBSMALLINT msgstate,
               DBSMALLINT severity, char *msgtext);
int remoteerr(DBPROCESS *dbproc, int severity, int dberr, int oserr,
              char *dberrstr, char *oserrstr);

int chk_err(SRV_SERVER *server, SRV_PROC *srvproc, int srverror,
            BYTE severity, BYTE state, int oserrnum, DBCHAR *errtext,
            int errtextlen, DBCHAR *oserrtext, int oserrtextlen);

void main(int argc, char **argv);


void main(argc, argv)
int argc;
char *argv[];
{
    SRV_CONFIG *config; // The configuration structure
    SRV_SERVER *server; // The service process
    int exitcode;

    // Allocate a configuration structure that is used to initialize
    // the Open Data Services application
    //
    config = srv_config_alloc();

    // Do server config steps here, for example, specify the log file
    //
    srv_config(config, (DBINT)SRV_LOGFILE, "gateway.log", SRV_NULLTERM);

    // Install the Open Data Services error handler.
    //
    srv_errhandle(chk_err);

    // Initialize the gateway and save the server handle
    // so it can be used in later functions.
    //
    server = srv_init(config, "GATEWAY", SRV_NULLTERM);
    if (server == NULL) {
        printf("Unable to initialize Gateway.\n");
        ExitProcess(exitcode);
    }

    // When starting the gateway, initialize the remote server structure.
    // This is done in the init_server() function.
    // All the other event handlers are also defined in the init_server()
    // function.
    //
    srv_handle(server, (DBINT)SRV_START, init_server);

    // Now everything's ready to go with our gateway, so we
    // start it and keep it going until we get a stop request.
    //
    srv_log(server, TRUE, "Gateway Starting", SRV_NULLTERM);
	
    srv_run(server);

}


//
// INIT_SERVER
//     Initialize the server on a SRV_START event.
//     Event handlers for the server are installed.
//
// Parameters:
//     server - Pointer to SRV_SERVER structure
//
// Returns:
//     SRV_CONTINUE
//
int init_server(server)
SRV_SERVER *server;
{
    char log_buffer[256];

	// server has just been started.  first thing to do is install all the other event
	// handlers
    srv_handle(server, (DBINT)SRV_CONNECT, init_remote);
    srv_handle(server, (DBINT)SRV_LANGUAGE, lang_execute);
    srv_handle(server, (DBINT)SRV_RPC, rpc_execute);
    srv_handle(server, (DBINT)SRV_DISCONNECT, exit_remote);
    srv_handle(server, (DBINT)SRV_ATTENTION, attn_handler);

	 // Initialize dblib and setup its error and message handlers
	 //
	 dbinit();
    dberrhandle(remoteerr);
    dbmsghandle(remotemsgs);

	 // Create mutex to serialize init_remote
	 //
	 init_remote_mutex = CreateMutex(NULL, FALSE, NULL);

     return(SRV_CONTINUE);
}

// INIT_REMOTE
//    Event handler for a SRV_CONNECT event.
//    A connection is made to the remote DBMS.
//
// Parameters:
//      srvproc - the handle to the client connection that got the SRV_CONNECT.
//
// Returns:
//     SRV_CONTINUE
//
// Side Effects:
//     If the connection to the remote dbms cannot be made, then issue
//     a SRV_DISCONNECT request.
//
//
int init_remote(srvproc)
SRV_PROC *srvproc;
{
	 char *string;
	 int len;
    REMOTE_DBMS *remote;    // the private data area to keep
                                // track of the remote DBMS
                                // connection.

    // This function is not reentrant, becuase dbopen() is not re-entrant
	// So first get exclusive access.

	WaitForSingleObject(init_remote_mutex, INFINITE);
	
	// Also we need to set a global variable Newsrvproc used by the dblib error
	// handler if it gets an error on dbopen().
        
    Newsrvproc = srvproc;

	// Allocate a REMOTE_DBMS information structure.
    //
    remote = (REMOTE_DBMS *)srv_alloc((DBINT)sizeof(*remote));

    // Allocate the LOGINREC structure used to make connections to the
    // remote server. Open the connection in the SRV_CONNECT handler.
    //
    remote->dbproc = (DBPROCESS *)NULL;

    remote->login = dblogin();

    // Set the user name, password, and application name for the remote DBMS.
    //
    DBSETLUSER(remote->login, srv_pfield(srvproc, SRV_USER, (int *)NULL));
    DBSETLPWD(remote->login, srv_pfield(srvproc, SRV_PWD, (int *)NULL));
    DBSETLAPP(remote->login, srv_pfield(srvproc, SRV_APPLNAME, (int *)NULL));
    DBSETLNATLANG(remote->login, srv_pfield(srvproc, SRV_NATLANG, (int *)NULL));

    // Try to open a connection to the remote DBMS.
    //
    if ((remote->dbproc = dbopen(remote->login,
                                 "myserver")) == (DBPROCESS *)NULL) {

			// Send a message to the client that
			// the remote connection failed.
			//
			srv_sendmsg(srvproc, 
					SRV_MSG_ERROR, 
					(DBINT)REMOTE_FAIL, 
					(DBTINYINT)0,
					(DBTINYINT)0, 
					NULL, 
					0, 
					0, 
					"Login to remote DBMS failed (dbopen).",
					SRV_NULLTERM);

        // Deallocate the data structures
        //
		dbfreelogin(remote->login);
        srv_free(remote);
        srv_setuserdata(srvproc, (BYTE *)NULL);
		ReleaseMutex(init_remote_mutex);
        return(SRV_DISCONNECT);

    } else {
        // Connection to the remote DBMS successful.  Save
        // remote data structure in the "srvproc" so it will be
        // available to the other handlers. We'll also map the remote
        // DBMS connection to our "srvproc".
        //
        srv_setuserdata(srvproc, (VOID *)remote);
        dbsetuserdata(remote->dbproc, (VOID *)srvproc);
    }

	 ReleaseMutex(init_remote_mutex);
    
    return(SRV_CONTINUE);
}

//
// LANG_EXECUTE
//     Execute a client language request on the remote dbms.
//     Any results from the remote dbms are passed back to the client.
//
//  Parameters:
//     srvproc -  process handle to the current client connection.
//
// Returns:
//     SRV_CONTINUE
//
int lang_execute(srvproc)
SRV_PROC *srvproc;
{
    REMOTE_DBMS *rmt_dbms;  // the remote database pointer
    DBPROCESS *rmtproc; // our DBPROCESS pointer
    DBCHAR *query;      // pointer to language buffer
    int query_len;          // length of query
    int status;             // status of dblib calls

    // Get the remote dbms pointer we saved in the srvproc via
    // srv_setuserdata.
    //
    rmt_dbms = (REMOTE_DBMS *)srv_getuserdata(srvproc);

    // Get the pointer to the remote DBPROCESS
    //
    rmtproc = rmt_dbms->dbproc;

    // Get the pointer to the client language request command buffer.
    //
    query = srv_langptr(srvproc);

    // Place buffer into target SQL server's buffer.
    status = dbcmd(rmtproc, query);
    
    // If previous DBLIB call successful, send command buffer to SQL Server.
    //
    if (status == SUCCEED) 
        status = dbsqlexec(rmtproc);

    if (!SRV_GOT_ATTENTION(srvproc) && (status == SUCCEED)) {
        //
        // Get the remote dbms results and pass them back to
        // client.
        //
        handle_results(rmtproc, srvproc);
    } else {
        //
        // If an attention event was issued or the dbsqlexec failed,
        // acknowledge with senddone.  Any backend error message would 
		// have already been sent by the dblib error handler.
        //
            srv_senddone(srvproc, SRV_DONE_FINAL, (DBUSMALLINT)0, (DBINT)0);
    }
    return(SRV_CONTINUE);
}

//
// RPC_EXECUTE
//     Execute a client RPC.
//	   Uses the dbrpc* to send requests on to SQL Server.
//
// Parameters:
//    srvproc - The Open Server process handle to use to send results to the
//               client.
//
// Returns:
//     SRV_CONTINUE
//
int rpc_execute(srvproc)
SRV_PROC *srvproc;
{
    DBPROCESS *rmtproc; // Our DBPROCESS pointer
    int i;                  // Index variable
    short params;
    short retparam;
    short *paramarray;
	 DBCHAR complete_rpc_name[4 * (MAXNAME + 1)]; //database.owner.name;#
	 DBCHAR *rpc_db;
	 DBCHAR *rpc_owner;
	 DBCHAR *rpc_name;
	 int rpc_number;
	 char rpc_number_char[17];
	 DBCHAR *rpc_paramname;
	 BYTE rpc_paramstatus;
	 int rpc_paramtype;
	 DBINT rpc_parammaxlen;
	 DBINT rpc_paramlen;
	 void *rpc_paramdata;

    rmtproc = ((REMOTE_DBMS *)srv_getuserdata(srvproc))->dbproc;
    paramarray = ((REMOTE_DBMS *)srv_getuserdata(srvproc))->retparams;

   // Construct full RPC name and initialize the RPC to the remote DBMS.
   //
	memset (complete_rpc_name, 0, sizeof(complete_rpc_name));
	rpc_db = srv_rpcdb(srvproc, (int *)NULL);
	rpc_owner = srv_rpcowner(srvproc, (int *)NULL);
	rpc_name = srv_rpcname(srvproc, (int *)NULL);
	
	if (rpc_db != NULL) {
		strcat(complete_rpc_name,rpc_db);
		strcat(complete_rpc_name,".");
	};
	if (rpc_owner != NULL) {
		strcat(complete_rpc_name,rpc_owner);
		strcat(complete_rpc_name,".");
	};
	strcat(complete_rpc_name,rpc_name);

	rpc_number = srv_rpcnumber(srvproc);
	if (rpc_number > 0) {
    	strcat(complete_rpc_name,";");
		itoa(rpc_number,rpc_number_char,10);
		strcat(complete_rpc_name, rpc_number_char);
	};
	
	dbrpcinit(rmtproc, complete_rpc_name, (DBUSMALLINT)srv_rpcoptions(srvproc));

    // Set up any RPC parameters.
    //
    params = srv_rpcparams(srvproc);
    retparam = 1;

    for (i = 1; i <= params; i++) {
		rpc_paramname = srv_paramname(srvproc, i, (int *)NULL);
	    if (!strlen(rpc_paramname)) 
			 rpc_paramname = NULL;
       rpc_paramstatus = (BYTE)srv_paramstatus(srvproc, i);
       rpc_paramtype = srv_paramtype(srvproc, i);
       rpc_parammaxlen = srv_parammaxlen(srvproc, i);
		rpc_paramlen = srv_paramlen(srvproc, i);
       rpc_paramdata = srv_paramdata(srvproc, i);
		dbrpcparam(rmtproc,rpc_paramname,rpc_paramstatus,rpc_paramtype, 
			 rpc_parammaxlen,rpc_paramlen, rpc_paramdata);

        // The current rpc may have three parameters, but only the
        // first and third are return parameters.  This means that
        // dbnumrets() returns two parameters, not three.  The first
        // call to dbretdata() referes to the rpc's first parameter,
        // and the second call to dbretdata() refers to the rpc's
        // third parameter.  To handle this, we map each return
        // parameter to its original parameter so we can later
        // reset the return value of the correct return parameters
        // in "handle_results()".
        //
        if ((BYTE)srv_paramstatus(srvproc, i) & SRV_PARAMRETURN) {
            paramarray[retparam] = i;
            retparam++;
        }
    }

    // Send the RPC to the remote DBMS.
    //
    dbrpcsend(rmtproc);
    dbsqlok(rmtproc);

    // Now get the remote DBMS results and pass them back to
    // Open Server client.
    //
    handle_results(rmtproc, srvproc);
    return(SRV_CONTINUE);
}

//
// HANDLE_RESULTS
//     Once a command has been sent to the remote DBMS by the
//     SRV_LANGUAGE  handler, this routine processes the results
//     and passes them back to the client.
//
// Parameters:
//     rmtproc - The DBPROCESS handle to the remote DBMS.
//     srvproc - The process handle to use to send results to the client.
//
// Returns:
//     SUCCEED or FAIL.
//
int handle_results(rmtproc, srvproc)
DBPROCESS *rmtproc;
SRV_PROC *srvproc;
{
    short cols;             // data columns returned
    int i;                  // index variable
    DBINT rows;             // number of rows sent
    BOOL results_sent;      // were result sets sent
    
    results_sent = FALSE;
    rows = 0L;
    
    // Process the results from the remote DBMS.
    // Since a command may consist of multiple commands or a single
    // command that has multiple sets of results, we'll loop through
    // each results set.
    //
    while (TRUE) {

        returnvalue = dbresults(rmtproc);
        if (returnvalue == NO_MORE_RESULTS) {
            break;
        }

        // Check to see if the client has sent an attention event.  If
        // so, simply discard data from remote server
        //
        if (SRV_GOT_ATTENTION(srvproc)) {
    		dbcancel(rmtproc);	// See attn_handler comments
        	continue;
        }

        //
        // If this is the second time through the loop,
        // send a completion message to the client
        // for the previous results sent.
        //
        if (results_sent == TRUE) {

            // If the previous batch was one that may
            // have returned rows, set the DONE status
            // accordingly.
            //
            if (rows > 0) {
                srv_senddone(srvproc, SRV_DONE_MORE | SRV_DONE_COUNT,
                             (DBUSMALLINT)0, rows);
            } else
                srv_senddone(srvproc, SRV_DONE_MORE, (DBUSMALLINT)0,
                             (DBINT)0);
        }

        // How many data columns are in the row?
        // Non-"select" statements will have 0 columns.
        //
        cols = dbnumcols(rmtproc);

        // Build the row description for the client return.
        //
        for (i = 1; i <= cols; i++) {

            srv_describe(srvproc, i, dbcolname(rmtproc, i), SRV_NULLTERM,
                         (DBINT)dbcolntype(rmtproc, i), dbcollen(rmtproc, i),
                         (DBINT)dbcolntype(rmtproc, i), dbcollen(rmtproc, i),
                         (BYTE *)NULL);
        }

        // Send each row from the remote DBMS to the client.
        //
        rows = 0;
        while (TRUE) {

            returnvalue = dbnextrow(rmtproc);
            if (returnvalue == NO_MORE_ROWS) {
                break;
            }

            // The row description is built.  Move the
            // rows from the remote server to the client.
            //
            for (i = 1; i <= cols; i++) {
                srv_setcollen(srvproc, i, (short)dbdatlen(rmtproc, i));
                srv_setcoldata(srvproc, i, dbdata(rmtproc, i));

            // Check to see if the client has issued an attention event.
            // If so, discard data from the remote server.
            //
            if (SRV_GOT_ATTENTION(srvproc)) {
    			dbcancel(rmtproc);	// See attn_handler comments
                continue;
            }
            if (srv_sendrow(srvproc) == SUCCEED)
                rows++;
        }  // end of rows loop

        // If the command was one where count is meaningful, DBCOUNT()
        // will be >= 0 and we will set rows accordingly.
        //
        if (dbnumcols(rmtproc) == 0)
            rows = DBCOUNT(rmtproc);

        // Set flag so that we will send a completion
        // message for the current set of results.
        //
        results_sent = TRUE;
    } // end of results loop


    // Send the final done packet for the execution of the command batch.
    //
    //
    if (rows > 0) {
        srv_senddone(srvproc, SRV_DONE_COUNT | SRV_DONE_FINAL, (DBUSMALLINT)0,
                     rows);
    } else {
        srv_senddone(srvproc, SRV_DONE_FINAL, (DBUSMALLINT)0, (DBINT)0);
    }
    return(SUCCEED);
}

// EXIT_REMOTE
//     Handler for SRV_DISCONNECT events.
//     Closes remote DBMS connection if appropriate.
//
// Parameters:
//     srvproc - the handle to the client connection
//
// Returns:
//     SRV_CONTINUE
//
int exit_remote(srvproc)
SRV_PROC *srvproc;
{
	 char *string;
	 int len;
    REMOTE_DBMS *remote;    // pointer to target connect structure

    remote = (REMOTE_DBMS *)srv_getuserdata(srvproc);

    // Is there a REMOTE_DBMS structure to clean-up?
    //
    if (remote != (REMOTE_DBMS *)NULL) {

        // Is there a live dbproc?
        //
        if (remote->dbproc != (DBPROCESS *)NULL) {
            dbclose(remote->dbproc);
        }
        dbfreelogin(remote->login);
        srv_free(remote);
    }

    return(SRV_CONTINUE);
}

// CHK_ERR
//     Handles ODS errors by logging them or sending to client
//
// Parameters:
//     as documented in the Open Data Services Programmer's Reference
//
// Returns:
//     SRV_CONTINUE, SRV_CANCEL, or SRV_EXIT
//

RETCODE chk_err(server, srvproc, errornum, severity, state, oserrnum, errtext,
                errtextlen, oserrtext, oserrtextlen)
SRV_SERVER *server;
SRV_PROC *srvproc;
int errornum;
BYTE severity;
BYTE state;
int oserrnum;
DBCHAR *errtext;
int errtextlen;
DBCHAR *oserrtext;
int oserrtextlen;
{
    char log_buffer[256];
    char error[256];
    char oserror[256];

    memcpy(error, errtext, errtextlen);
    error[errtextlen] = '\0';
    memcpy(oserror, oserrtext, oserrtextlen);
    oserror[oserrtextlen] = '\0';

    // The text strings must be freed, because allocated by FormatMessage
    //
    LocalFree(errtext);
    LocalFree(oserrtext);

	// Strip out resource information. Get the actual error number.
	errornum = (errornum & 0x0000FFFF);

    // Operating system error?
    //
    if (oserrnum != SRV_ENO_OS_ERR) {
        sprintf(log_buffer, "SERVER OS ERROR: %d: %s.", oserrnum, oserror);

        if (server)	
			srv_log(server, TRUE, log_buffer, SRV_NULLTERM);
		else	// If application not initialized log to screen
    		printf ("%s\n", log_buffer);
    }

    // Is this a fatal error for the server?
    //
    if (severity >= SRV_FATAL_SERVER) {
        sprintf(log_buffer, 
		  			"SERVER: FATAL SERVER ERROR: errornum = %d, severity = %d, state = %d: %s.",  
					 errornum, severity, state, error);
        if (server)	
			srv_log(server, TRUE, log_buffer, SRV_NULLTERM);
		else	// If application not initialized log to screen
    		printf ("%s\n", log_buffer);
        return(SRV_EXIT);
    } else {
        //
        // Did the "srvproc" get a fatal error?
        //
        if (severity >= SRV_FATAL_PROCESS) {
            sprintf(log_buffer,
						  "SERVER: FATAL CONNECT ERROR: errornum = %d, severity = %d, state = %d: %s.",
                     errornum, severity, state, error);
        	if (server)	
				srv_log(server, TRUE, log_buffer, SRV_NULLTERM);
			else	// If application not initialized log to screen
    			printf ("%s\n", log_buffer);	
            return(SRV_CANCEL);
        }
    }

    // A non-fatal error or an information message received.
    // We'll pass it through to the client.
    //
    if (srvproc != (SRV_PROC *)NULL && (server != NULL))
        if (severity < 10) {    // if informational message
            srv_sendmsg(srvproc, SRV_MSG_INFO, (DBINT)errornum, severity, 0,
                        NULL, 0, 0, error, SRV_NULLTERM);
        } else {            // must be an error message
        srv_sendmsg(srvproc, SRV_MSG_ERROR, (DBINT)errornum, severity, 0,
                    NULL, 0, 0, error, SRV_NULLTERM);
    } else {
        sprintf(log_buffer, "ODS ERROR: errornum = %d, severity = %d: %s",
                errornum, severity, error);
        if (server)	
			srv_log(server, TRUE, log_buffer, SRV_NULLTERM);
		else	// If application not initialized log to screen
    		printf ("%s\n", log_buffer);	
    }
    return(SRV_CONTINUE);
}

// REMOTEMSGS
//     DBLIB calls this routine when any messages are received
//     from the remote DBMS. It gets the remote message information and
//     sends it back to the client as a message.
//
// Parameters:
//     as documented in SQL Server Programmer's Reference for C
//
// Returns:
//     0
//

int remotemsgs(dbproc, msgno, msgstate, severity, msgtext)
DBPROCESS *dbproc;
DBINT msgno;
DBSMALLINT msgstate;
DBSMALLINT severity;
char *msgtext;
{
    SRV_PROC *srvproc;

    // If a remote DBMS error was received during the remote
    // open, the dbproc is NULL and a message is sent back on the
    // most recent srvproc.
    //
    if (dbproc == (DBPROCESS *)NULL) {
        srvproc = Newsrvproc;
    } else {
        if ((srvproc = (SRV_PROC *)dbgetuserdata(dbproc)) == NULL) {
            //
            // An error was received after the dbproc was assigned, but
            // before we were able to associate our srvproc.
            //
            srvproc = Newsrvproc;
        }
    }
    if (severity < 10) {    // if informational message

        srv_sendmsg(srvproc, SRV_MSG_INFO, msgno, (DBTINYINT)severity,
                    (DBTINYINT)msgstate, NULL, 0, 0, msgtext, SRV_NULLTERM);
    	return(0);
    } 

	// Trap login fail message
	if (msgno == REMOTE_FAIL){
        // Send a message to the client that
        // the remote connection failed.
        //
        srv_sendmsg(srvproc, SRV_MSG_ERROR, (DBINT)msgno, (DBTINYINT)severity,
                    (DBTINYINT)msgstate, NULL, 0, 0,
                    "Login to remote DBMS failed (dbopen).", SRV_NULLTERM);
	}

	// must be an error message
    srv_sendmsg(srvproc, SRV_MSG_ERROR, msgno, (DBTINYINT)severity,
    	(DBTINYINT)msgstate, NULL, 0, 0, msgtext, SRV_NULLTERM);

    return(0);
}

// REMOTEERR
//     This is the handler for error messages from the remote DBMS, in this
//     case DBLIB.  It gets called whenever a DBLIB error occurs.  It takes
//     the error message and passes it back to the client.
//
// Parameters:
//     as documented in SQL Server Programmer's Reference for C
//
// Returns:
//     INT_EXIT to exit the program.
//     INT_CANCEL to cause a FAIL return from the DBLIB routine that got
//     the error.
//
int remoteerr(dbproc, severity, dberr, oserr, dberrstr, oserrstr)
DBPROCESS *dbproc;
int severity;
int dberr;
int oserr;
char *dberrstr;
char *oserrstr;
{
    SRV_PROC *srvproc = (SRV_PROC *)NULL;

    // If the DBLIB process is dead or we get a DBLIB error SQLESMSG
	// ("General SQL Server Error:...") then simply ignore it.  The error
    // message has already been sent to the client.
    //
    if (dberr == SQLESMSG) {
        return(INT_CANCEL);
    }

    //
    // A remote DBMS error may have been issued during the remote
    // open. In this case, the dbproc will be NULL and a message
    // will be sent on the most recent srvproc.
    //
    if (dbproc == (DBPROCESS *)NULL) {
        srvproc = Newsrvproc;
    } else {
        if ((srvproc = (SRV_PROC *)dbgetuserdata(dbproc)) == NULL) {
            // An error was issued after the dbproc was assigned but before
            // we were able to associate our srvproc.
            //
            srvproc = Newsrvproc;
        }
    }

	// Trap connection failure error
	if (dberr == SQLECONN){
        // Send a message to the client that
        // the remote connection failed.
        //
        srv_sendmsg(srvproc, SRV_MSG_ERROR, (DBINT)REMOTE_FAIL, (DBTINYINT)severity,
                    (DBTINYINT)0, NULL, 0, 0,
                    "Unable to establish connection to remote DBMS (dbopen).", SRV_NULLTERM);

        return(INT_CANCEL);
	}

    //
    // Send error message to client.
    //
    srv_sendmsg(srvproc, SRV_MSG_ERROR, (DBINT)REMOTE_MSG,
                (DBTINYINT)severity, (DBTINYINT)0, NULL, 0, 0, dberrstr,
                SRV_NULLTERM);

    if (oserr != DBNOERR) {
        srv_sendmsg(srvproc, SRV_MSG_ERROR, (DBINT)REMOTE_MSG,
                    (DBTINYINT)severity, (DBTINYINT)0, NULL, 0, 0, oserrstr,
                    SRV_NULLTERM);
    }
    return(INT_CANCEL);
}

// ATTN_HANDLER
//     This is an event handler that will be called when the gateway receives
//     an attention from a client.
//
// Parameters:
//     srvproc - Pointer to the client connection structure
//               receiving the attention.
//
// Returns:
//     SRV_CONTINUE
//
int attn_handler(srvproc)
SRV_PROC *srvproc;
{
	// Open Data Services NT receives client attention events asynchronously.
  	// Because handle_results may be calling dbresults or dbnextrow, we can not
	// process the attention with the dbcancel call here. Instead dbcancel
	// will be called after the attention has been detected using 
	// SRV_GOT_ATTENTION.

    return(SRV_CONTINUE);
}


