//
// This program is an example of an Open Data Services application. It accepts
// requests from clients to execute stored procedures either as language
// events or as remote stored procedure calls.
// It may also be invoked using the Service Control Manager.
//

// Event handlers for SP requests.
//
RETCODE proclist();         // List SP associated with Server
RETCODE diskfree();         // Return the space on a given drive
RETCODE disklist();         // List available drives and their space

// Stored Procedure parameter information structure.
//
typedef struct sp_params {
    DBCHAR name[MAXNAME];           // Parameter name
    int    type;                    // Parameter data type
    DBINT  length;                  // Parameter type length
    int    status;                  // Parameter return status
    DBCHAR defaultvalue[MAXLEN *10];// Optional default value
} SP_PARAMS;

// Stored Procedure information structure.
//
typedef struct sp_info {
    DBCHAR      name[MAXNAME];          // Procedure name
    DBCHAR      usage[MAXLEN];          // Usage string
    int         numparams;              // Number of parameters
    SP_PARAMS   params[MAXPARAMS];      // Parameter array
    RETCODE   (*handler)(VOID *, ...);  // Pointer to function with variable arguments
} SP_INFO;

// Array of Stored Procedure handlers.
//
SP_INFO Sps[] =
{
    "proclist",
    "usage: proclist",
    0,
       {
        0
       },
    proclist,

    "disklist",
    "usage: disklist",
    0,
   {
        0
    },
    disklist,

    "diskfree",
    "usage: diskfree <[@drive =] drive letter> [,] <[@space =] free space>",
    2,
   {
        "drive",
        SRVCHAR,
        1,
        VALUE,              // pass by value
        "c",

        "space",
        SRVINT4,
        8,
        REFERENCE,          // pass by reference
        "0",
    },
    diskfree,

};

#define Rpcnumber sizeof(Sps) / sizeof(SP_INFO)

// Other function prototypes etc.
RETCODE init_remote(SRV_PROC * srvproc);
RETCODE init_server(SRV_SERVER * server);
RETCODE sp_execute(SRV_PROC * srvproc);
RETCODE exit_remote(SRV_PROC *srvproc);
RETCODE lang_execute(SRV_PROC *srvproc);

// Miscellaneous prototypes for sp language events
//
RETCODE lang_parser(SRV_PROC *srvproc, SP_INFO **sp,
                    DBCHAR paramvalues[MAXPARAMS][MAXLEN]);

DBCHAR *scan_next(DBCHAR *string, DBCHAR *word);


// =================================== main ==================================
//
void main( int argc, char *argv[] )
{

    // Lots of great sample code here to install procsrv as a service.  
    // Eventually does same thing as gateway:  srv_init, srv_config, install startup handler
}
    


// ================================ lang_execute =============================
//
// LANG_EXECUTE
//    Execute a client language request on the procedure server.
//
// Parameters:
//    srvproc - process handle to the current client connection.
//
// Returns:
//    SRV_CONTINUE
//
RETCODE lang_execute( SRV_PROC *srvproc )
{
    int      i;
    DBCHAR   paramvalues[MAXPARAMS][MAXLEN];
    BYTE     convertvalues[MAXPARAMS][MAXLEN];
    SP_INFO *sp = NULL;

    // Initialize parameter storage
    //
    for (i = 0; i < MAXPARAMS; i++) {
        memset(paramvalues[i], 0, MAXLEN);
        memset(convertvalues[i], 0, MAXLEN);
    }

	// send the query string through language parser, comes back as filled-in 
	// sp_info structure
    if (lang_parser(srvproc, &sp, paramvalues) == SUCCEED) {
        for (i = 0; i < sp->numparams; i++) {
			// check to see if params match sp_info description, if not send error
            if (strlen(paramvalues[i]) == 0 &&
                strlen(sp->params[i].defaultvalue) == 0) {
                srv_sendstatus(srvproc, 1);
                srv_sendmsg(srvproc, SRV_MSG_ERROR, INVALID_SP_SYNTAX,
                            SRV_INFO, (DBTINYINT)0, NULL, 0, 0, sp->usage,
                            SRV_NULLTERM);
                srv_senddone(srvproc, (SRV_DONE_ERROR | SRV_DONE_FINAL), 0,
                             0);
                return SRV_CONTINUE;
            }
            if (strlen(paramvalues[i]) == 0 &&
                strlen(sp->params[i].defaultvalue) != 0)
                strcpy(paramvalues[i], sp->params[i].defaultvalue);

            // convert parameters from character string to parmeter type
            //
            srv_convert(srvproc, SRVCHAR, paramvalues[i], -1,
                        sp->params[i].type, &convertvalues[i],
                        sp->params[i].length);
        }

        // Execute the procedure
        //
        (sp->handler)((VOID *)srvproc, &convertvalues[0], &convertvalues[1],
                      &convertvalues[2], &convertvalues[3]);
    }
    return SRV_CONTINUE;
}




// ================================== sp_execute =============================
//
// SP_EXECUTE
//    Execute a client stored procedure.
//
//    Scans the list of defined stored procedures, checks the parameters and
//    executes the procedure. If results are returned it is the responsiblity
//    of the underlying proedure.
//
// Parameters:
//    srvproc - The process handle to use to send results to the client.
//
// Returns:
//    SRV_CONTINUE
//
RETCODE sp_execute( SRV_PROC  *srvproc )
{
    int      i;
    int      x;
    int      y;
    int      len;
    int      numparams;
    SP_INFO *sp = NULL;
    DBCHAR   msg[MAXLEN];
    DBINT    msgnum;
    DBCHAR   paramvalues[MAXPARAMS][MAXLEN];
    DBINT    paramtype;
    DBCHAR  *value;
    DBINT    type;

    numparams = srv_rpcparams( srvproc );

    for( y = 1; y <= numparams; y++ )
    {
        // Check parameters for correct type
        //
        paramtype = srv_paramtype( srvproc, y );

	    // compare with sp_info structure

        if( type != paramtype ) {
            sprintf( msg, "Parameter \'%d \' is incorrect type.", y );
            msgnum = BAD_SP_PARAMETER;
            goto error;
        }

        // other param checks dones here
        // Move SP parameters to paramvalues[] local variables
        
    }
    
    // Execute procedure
    //
    (*sp->handler)( (VOID *)srvproc,
                    paramvalues[0],
                    paramvalues[1],
                    paramvalues[2],
                    paramvalues[3] );

    return SRV_CONTINUE;

error:
    srv_sendstatus( srvproc, 1 );
    srv_sendmsg( srvproc, SRV_MSG_ERROR, msgnum, SRV_INFO, (DBTINYINT)0, NULL,
                 0, 0, msg, SRV_NULLTERM );
    srv_senddone( srvproc, (SRV_DONE_ERROR | SRV_DONE_FINAL), 0, 0 );
    return SRV_CONTINUE;
}



// The following are the supported store procedure functions
//
// ================================== proclist ===============================
//
// PROCLIST
//    Returns the usage for all defined stored procedures
//
// Parameters:
//    srvproc - the handle to the client connection that got the SRV_CONNECT.
//
// Returns:
//    SUCCEED
//
// Side Effects:
//    Returns a result set to client
//
RETCODE proclist( SRV_PROC *srvproc )
{
    DBCHAR colname1[MAXNAME];
    DBCHAR colname2[MAXNAME];
    int i;

    sprintf(colname1, "spname");
    srv_describe(srvproc, 1, colname1, SRV_NULLTERM, SRVCHAR, MAXNAME,
                 SRVCHAR, 0, NULL);

    sprintf(colname2, "spusage");
    srv_describe(srvproc, 2, colname2, SRV_NULLTERM, SRVCHAR, MAXLEN, SRVCHAR,
                 0, NULL);

    // Return each SP handler as a row
    //
    for (i = 0; i < Rpcnumber; i++) {
        srv_setcoldata(srvproc, 1, Sps[i].name);
        srv_setcollen(srvproc, 1, strlen(Sps[i].name));

        srv_setcoldata(srvproc, 2, Sps[i].usage);
        srv_setcollen(srvproc, 2, strlen(Sps[i].usage));

        srv_sendrow(srvproc);
    }
    srv_senddone(srvproc, (SRV_DONE_COUNT | SRV_DONE_FINAL), 0, i);

    return SUCCEED;
}



// =================================== disklist ==============================
//
// DISKLIST
//     Returns a row for each defined drive containing its name and the
//     amount of disk space available.
//
// Parameters:
//    srvproc - the handle to the client connection that got the SRV_CONNECT.
//
// Returns:
//    SUCCEED
//
// Side Effects:
//     Returns a result set to client
//
RETCODE disklist( SRV_PROC  *srvproc )
{
    DBCHAR colname1[MAXNAME];
    DBCHAR colname2[MAXNAME];
    DBCHAR drivename;
    DBCHAR rootname[16];
    int    drivenum;
    int    secPerCluster;
    int    bytesPerSector;
    int    freeClusters;
    int    totalClusters;
    int    drivenums;
    int    space_remaining;
    int    i = 0;

    sprintf( colname1, "drive" );
    srv_describe( srvproc, 1, colname1, SRV_NULLTERM, SRVCHAR, 1, SRVCHAR, 1,
                  (BYTE *)&drivename );

    sprintf( colname2, "bytes free" );
    srv_describe( srvproc, 2, colname2, SRV_NULLTERM, SRVINT4, 4, SRVINT4, 4,
                  (BYTE *)&space_remaining );

    drivenums = GetLogicalDrives();

    drivenums >>= 2;        // Ignore drives A and B
    for( drivename = 'C', drivenum = 3; drivename <= 'Z';
         drivename++, drivenum++ )
    {
        if( drivenums & 1 )
        {
            i++;

            sprintf( rootname, "%c:\\", drivename );
            GetDiskFreeSpace( rootname, &secPerCluster, &bytesPerSector,
                              &freeClusters, &totalClusters );

            space_remaining = secPerCluster * freeClusters * bytesPerSector;

            srv_sendrow( srvproc );
        }

        drivenums >>= 1;
    }

    srv_senddone( srvproc, (SRV_DONE_COUNT | SRV_DONE_FINAL), 0, i );
    return SUCCEED;
}


