//****************************************************************************
// xp_startmail
//      This is the extended procedure enable SQL Server for mail.
//      It spawns a thread to serially service all mail requests.
//      All SQL Server mail is serialized because the current version of MAPI
//      is single threaded.
//
// Parameters:
//    srvproc -  pointer to SRV_PROC stucture
//
// Returns:
//    XP_NOERROR | XP_ERROR
//
// Side Effects:
//      Spawns a mail service thread and defines a pre-stop handler
//****************************************************************************
RETCODE xp_startmail(srvproc)
SRV_PROC *srvproc;
{
    SRV_SERVER *server = srvproc->server;
    THREAD cthreadID;       // thread ID of connection thread
    HANDLE cthreadHDL = 0;      // thread ID of connection thread
    DBCHAR * message = NULL;
	 RETCODE status = XP_NOERROR;
	 BOOL * notify = NULL;

    // If already logged in then return.
    //
    if (MailSession.bLoggedOn) {
		// alloc message string, load from DLL, etc., 
		// then:
        srv_sendmsg(srvproc, SRV_MSG_INFO, ALREADY_MAIL_SESSION, 0, (DBTINYINT)0,
                    NULL, 0, 0, message, SRV_NULLTERM);
        srv_senddone(srvproc, SRV_DONE_MORE, 0, 0);
        return(XP_NOERROR);
    }


	 // get login params from rpc params for from Registry
	 //

	 // now create a mutex for exclusive access to mail functions, because not re-entrant
	 
    MailSession.hMailMutex = CreateMutex(NULL, FALSE, NULL);
    if (MailSession.hMailMutex == NULL)
        goto EXIT;

	// create two auto-reset events to communicate with domail thread
	// initial state is non-signaled

    MailSession.hPostMail = CreateEvent(NULL, FALSE, FALSE, NULL);
    if (MailSession.hPostMail == NULL)
        goto EXIT;

    MailSession.hSentMail = CreateEvent(NULL, FALSE, FALSE, NULL);
    if (MailSession.hSentMail == NULL)
        goto EXIT;

	// log message of mail session started
    srv_message_handler(NULL, 0x40000000L, 1, 0, 0, message, strlen(message),
                        NULL, 0);

    // Create Mail thread
    //
	if ((cthreadHDL = CreateThread(NULL, server->config.stacksize,
                                   (LPTHREAD_START_ROUTINE)domail, NULL,
                                   (DWORD)NULL, &cthreadID)) == NULL)
        goto EXIT;

	// Now wait for the thread to become active.
    //
	 while (WaitForSingleObject(MailSession.hSentMail, 1000) == WAIT_TIMEOUT){
		if (SRV_GOT_ATTENTION(srvproc)){
			srv_senddone(srvproc, SRV_DONE_MORE, 0, 0);
			return(XP_ERROR);
		}
	 }
	 // Check for login error. 
	 // If not logged on now, error was logged by MailLogon, 
	 // so now just send message to client.
	 //
  	 if (!MailSession.bLoggedOn)
		 goto EXIT2;

    // set pre-stop handler to shutdown mail
    //
    srv_pre_handle(server, NULL, SRV_STOP, stopmail, FALSE);

	// send message to client that startup succeeded
    srv_sendmsg(srvproc, SRV_MSG_INFO, STARTUP_SUCCEED, 0, (DBTINYINT)0, NULL,
                0, 0, message, SRV_NULLTERM);
    srv_senddone(srvproc, SRV_DONE_MORE, 0, 0);
    goto CLEANUP;


EXIT:// couldn't create mutex, events, or domail thread	
    // log failure cause 
    srv_message_handler(NULL, 0xC0000000L | STARTUP_FAIL, 18, 1, oserr,
                        message, strlen(message), oserrtext, oserrtextlen);
EXIT2:	// Send client failure message
    srv_sendmsg(srvproc, SRV_MSG_ERROR, STARTUP_FAIL2, 18, (DBTINYINT)1, NULL,
                0, 0, message, SRV_NULLTERM);
    srv_senddone(srvproc, SRV_DONE_ERROR | SRV_DONE_MORE, 0, 0);
    status = XP_ERROR;

CLEANUP:
	//release any resources used here, eg reg keys
	// then:
	 return(status);
}

//****************************************************************************
// stopmail
//      This is the pre-stop handler routine to shutdown SQL Server mail.
//      It signals the mail thread via a NULL-srvproc to logoff and exit.
//
// Parameters:
//    server - a pointer to SRV_SERVER stucture
//
// Returns:
//    SRV_CONTINUE
//
// Side Effects:
//
//****************************************************************************
RETCODE stopmail(server)
SRV_SERVER *server;
{

	// omitting error check for no mapi session

    WaitForSingleObject(MailSession.hMailMutex, INFINITE);

    MailSession.pMailMessage = NULL;
    SetEvent(MailSession.hPostMail);

    WaitForSingleObject(MailSession.hSentMail, INFINITE);
	ReleaseMutex(MailSession.hMailMutex);
    return(SRV_CONTINUE);
}


	//****************************************************************************
// xp_sendmail
// 	This is the extended procedure to send a mail message.
//  	It sets the global SRV_PROC and signals the mail thread of a pending
//    	send.
//
// Parameters:
//    srvproc - a pointer to SRV_SRVPROC stucture
//
// Returns:
//    XP_NOERROR | XP_ERROR
//
// Side Effects:
//
//****************************************************************************
RETCODE xp_sendmail(SRV_PROC *srvproc)
{
    MAIL_MESSAGE MailMessage;

    memset(&MailMessage, 0, sizeof(MAIL_MESSAGE));

    // check if the Server has logged into mail server
    //
    if (!MailSession.bLoggedOn){
        LoadString(MailSession.hinsCur, NO_MAIL_SESSION, message, MAXLEN - 1);
        srv_sendmsg(srvproc, SRV_MSG_ERROR, NO_MAIL_SESSION, 16, (DBTINYINT)1,
                    NULL, 0, 0, message, SRV_NULLTERM);
        srv_senddone(srvproc, SRV_DONE_ERROR | SRV_DONE_MORE, 0, 0);
        return(XP_ERROR);
    }

	// get exclusive control of mail session.  while waiting, check every
	// so often for a cancel from the user
	 while (WaitForSingleObject(MailSession.hMailMutex, 500) == WAIT_TIMEOUT){
		if (SRV_GOT_ATTENTION(srvproc)){
			srv_senddone(srvproc, SRV_DONE_MORE, 0, 0);
			return(XP_ERROR);
		}
	 }

	// Got control of mail session.	
    // Set srvproc and signal mail thread to send query
	MailMessage.pSrvProc = srvproc;
	MailSession.RetCode = XP_NOERROR;
    MailSession.pMailMessage = &MailMessage;
    SetEvent(MailSession.hPostMail);

	// now wait for domail+sendmail to say it is done.
	// again, check for cancels while waiting
	while (WaitForSingleObject(MailSession.hSentMail, 500) == WAIT_TIMEOUT){
		if (SRV_GOT_ATTENTION(srvproc)){
			srv_senddone(srvproc, SRV_DONE_MORE, 0, 0);
			MailMessage.bNoOutput = TRUE;
			ReleaseMutex(MailSession.hMailMutex);
			return(XP_ERROR);
		}
	 }

	// message sent, now let someone else in.
    ReleaseMutex(MailSession.hMailMutex);
    return(MailSession.RetCode);
}
//****************************************************************************
//
// domail
//		This is the main function of the spawned mail thread. It logins
//    Microsoft Mail and serially services mail requests.  It supports
//    	the xp_sendmail extended procedure.
//
// Parameters: none
//    
// Returns:
//
// Side Effects:
//
//****************************************************************************
void domail
{
    
	// Since this is a new thread, need to put in structured exception
	//  handling here  (not shown)

    // Call the mail util dll function.  hMsg will be modified:
    // Handler to mapi session will be assigned, flags will be set,
    // originator name will be initialized, etc
    if (MailLogon(&MailSession) == SUCCEED){
		// log success of login, 
		// then:
		 MailSession.bLoggedOn = TRUE;
		 SetEvent(MailSession.hSentMail);		 
    } 
    else {
    	// login didn't succed, so signal and exit domail thread
    	SetEvent(MailSession.hSentMail);  		 
	    return;
	}

    // Wait to process mail command
    while (TRUE) {
        WaitForSingleObject(MailSession.hPostMail, INFINITE);
    
		// a PostMail signal + a null MailMessage is set by stopmail
        if (MailSession.pMailMessage == NULL)			
            goto SHUTDOWN;

		// call the sendmail function
	    if (sendmail(&MailSession, MailSession.pMailMessage) == FAIL &&
				  MailSession.RetCode == XP_NOERROR)
				  MailSession.RetCode = XP_ERROR;

		if (MailSession.RetCode == XP_NOERROR && !MailSession.pMailMessage->bNoOutput) {
				// tell client that message mail successfully sent
				srv_sendmsg(MailSession.pMailMessage->pSrvProc, SRV_MSG_INFO, 0, 0, (DBTINYINT)0, NULL,
										0, 0, message, SRV_NULLTERM);
			  }
		  }

		  // Return status code & done to client
		  //
		  srv_sendstatus (MailSession.pMailMessage->pSrvProc, MailSession.RetCode);

		  if (MailSession.RetCode == XP_NOERROR)
				srv_senddone(MailSession.pMailMessage->pSrvProc, SRV_DONE_MORE,
					0, 0);
		  else 	// XP_ERROR
			  srv_senddone(MailSession.pMailMessage->pSrvProc,
					SRV_DONE_ERROR | SRV_DONE_MORE, 0, 0);
 
        // Let client thread know of completion
        SetEvent(MailSession.hSentMail);

    }  // end of continuous while loop

SHUTDOWN:
    // Log off and cleanup requested by stopmail
    //
    MailLogoff(&MailSession);
	// log mapi shutdown
	// then: 
	MailSession.bLoggedOn = FALSE;
	SetEvent(MailSession.hSentMail);
}

